

#include <avr/io.h>
#include <avr/interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>

#define SONG_LENGTH	200
#define UNKNOWN_NOTE	42
#define PAUSE_NOTE	20


struct sound {
	uint8_t note;
	uint32_t count;
};

static uint8_t frecv[8] = {119, 106, 94, 89, 79, 70, 63, 59};
static unsigned int rec_pressed = 0, play_pressed = 0;
static unsigned int rec_pos = 0, play_pos = 0;
static enum {PLAY, SING, RECORD} state = SING;

static struct sound rec_song[SONG_LENGTH];
static struct sound play_song[SONG_LENGTH];

static void init()
{
	/* Set the input pins for buttons and output for the LED. */
	DDRC = 0x00;
	DDRA = _BV(PA6);

	PORTC = 0xFF;
	PORTA = ~(_BV(PA6));

	/* Set the output pin for buzzer. */
	DDRD |= _BV(PD5);
	TCCR1A |= _BV(WGM10) | _BV(COM1A0);
	TCCR1B |= _BV(WGM13) | _BV(CS12);

	/* Setting CTC mode. */
	TCCR2A = _BV(WGM21);

	/* Setting the prescaler. */
	TCCR2B |= _BV(CS22) | _BV(CS21) | _BV(CS20);

	/* Setting the OCR; every 10 ms. */
	OCR2A = 156;

	/* Setting timer interrupts. */
	TIMSK1 |= (1 << OCIE1A);
}

static void inline set_pwm()
{
	TCCR1A |= _BV(COM1A0);
}

static void inline unset_pwm()
{
	TCCR1A &= ~_BV(COM1A0);
}

static void inline set_note(uint8_t f)
{
	OCR1A = f;
}


static void clear_song(struct sound song[])
{
	int i;

	for (i = 0; i < SONG_LENGTH; i ++) {
		song[i].note = UNKNOWN_NOTE;
		song[i].count = 0;
	}
}

static void copy_song(struct sound source[], struct sound dest[])
{
	int i;

	for (i = 0; i < SONG_LENGTH; i ++) {
		dest[i].note = source[i].note;
		dest[i].count = source[i].count;
	}
}

ISR(TIMER1_COMPA_vect)
{
	uint8_t note = 0;
	uint8_t plays = 0;

	/* Check if the RECORD button is pressed. Only change the state if it
	 * is pressed and it wasn't pressed at the last interupt.
	 *
	 * If the current state is RECORD, change it to SING (end recording),
	 * otherwise change the state to RECORD. */
	if (!(PINA & _BV(PA4))) {
		if (!rec_pressed) {
			if (state == RECORD) {
				state = SING;
			}
			else {
				state = RECORD;

				clear_song(rec_song);
				rec_pos = 0;
			}
		}
		rec_pressed = 1;
	}
	else {
		rec_pressed = 0;
	}

	/* Same thing for the PLAY button. */
	if (!(PINA & _BV(PA5))) {
		if (!play_pressed) {
			if (state == PLAY) {
				state = SING;
			}
			else {
				state = PLAY;

				play_pos = 0;
				copy_song(rec_song, play_song);
			}
		}
		play_pressed = 1;
	}
	else {
		play_pressed = 0;
	}

	/* Light LED on PA6 if the current state is record. Turn it off
	 * otherwise. */
	if (state == RECORD) {
		PORTA |= _BV(PA6);
	}
	else {
		PORTA &= ~_BV(PA6);
	}

	/* See if one of the buttond from port C (notes) is pressed. */
	for (note = 0; note < 8; note ++) {
		if (!(PINC & _BV(note))) {
			set_note(frecv[note]);
			set_pwm();
			plays = 1;

			break;
		}
	}

	/* No note is played. */
	if (plays == 0) {
		unset_pwm();
		note = PAUSE_NOTE;
	}

	/* If the state is record, remember the song. */
	if (state == RECORD && rec_pos < SONG_LENGTH - 1) {
		if (note == rec_song[rec_pos].note) {
			rec_song[rec_pos].count ++;
		}
		else {
			rec_song[++rec_pos].note = note;
			rec_song[rec_pos].count = 1;
		}
	}

	/* Go back to SING state if the memory is full. */
	if (state == RECORD && rec_pos >= SONG_LENGTH - 1) {
		state = SING;
	}

	/* If the state is PLAY, set the notes stored in the time buffer. */
	if (state == PLAY && play_pos <= rec_pos) {
		if (play_song[play_pos].count == 0) {
			play_pos ++;
		}

		if (play_pos > rec_pos) {
			unset_pwm();
			state = SING;
			return;
		}

		if (play_song[play_pos].note == PAUSE_NOTE) {
			unset_pwm();
		}
		else {
			set_note(frecv[play_song[play_pos].note]);
			set_pwm();
		}

		play_song[play_pos].count --;
	}
}

void main_loop()
{
	set_note(frecv[0]);
	set_pwm();

	while (1) {
		/* Forever and ever. */
	}
}

int main()
{
	sei();
	init();

	main_loop();

	return 0;
}
